import traceback
from datetime import datetime
from http import HTTPStatus

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.utils.timezone import make_aware


def get_log_data(
    timestamp,
    request: HttpRequest,
    response: HttpResponse | None,
    exception: Exception | None = None,
) -> dict:
    data = {
        "method": _get_method(request),
        "path": _get_path(request),
        "user_agent": _get_user_agent(request),
        "querystring": _get_querystring(request),
        "request_body": _get_request_body(request),
        "timestamp": _get_timestamp(timestamp),
        "server_ip": _get_server_ip(request),
        "client_ip": _get_client_ip(request),
        "status_code": (
            # if response is None, return status code 500
            _get_status_code(response) if response else HTTPStatus.INTERNAL_SERVER_ERROR
        ),
    }

    if exception:
        exception_data = {
            "exception_type": _get_exception_type(exception),
            "exception_message": str(exception),
            "traceback": _get_traceback(exception),
        }
        data.update(exception_data)

    return data


def _get_method(request: HttpRequest):
    return request.method


def _get_path(request: HttpRequest):
    return request.path


def _get_status_code(response: HttpResponse):
    return response.status_code


def _get_user_agent(request: HttpRequest):
    return request.META.get("HTTP_USER_AGENT", None)


def _get_querystring(request: HttpRequest):
    return (
        None
        if request.META.get("QUERY_STRING", None) == ""
        else request.META.get("QUERY_STRING", None)
    )


def _get_request_body(request: HttpRequest):
    return request.body.decode("utf-8") if request.body else None


def _get_timestamp(unix_timestamp: float) -> datetime:
    return (
        make_aware(datetime.fromtimestamp(unix_timestamp))
        if settings.USE_TZ
        else datetime.fromtimestamp(unix_timestamp)
    )


def _get_exception_type(exception: Exception) -> str:
    return exception.__class__.__name__


def _get_traceback(exception: Exception) -> str:
    return "".join(traceback.format_tb(exception.__traceback__))


def _get_server_ip(request: HttpRequest):
    return request.get_host()


def _get_client_ip(request: HttpRequest):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip



# Generated by Django 5.1.1 on 2024-10-07 14:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_logbox", "0002_remove_serverlog_duration"),
    ]

    operations = [
        migrations.RenameField(
            model_name="serverlog",
            old_name="exception",
            new_name="exception_type",
        ),
        migrations.AddField(
            model_name="serverlog",
            name="exception_message",
            field=models.TextField(null=True, verbose_name="exception_message"),
        ),
    ]



from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from ua_parser import Result, parse

from .models import ServerLog


@admin.register(ServerLog)
class ServerLogAdmin(admin.ModelAdmin):
    list_display = (
        "method",
        "path",
        "status_code",
        "user_agent_summary",
        "timestamp",
        "server_ip",
        "client_ip",
    )

    readonly_fields = (
        "method",
        "path",
        "status_code",
        "user_agent",
        "querystring",
        "request_body",
        "timestamp",
        "exception_type",
        "exception_message",
        "traceback",
        "server_ip",
        "client_ip",
    )

    fieldsets = (
        (
            _("Request Information"),
            {
                "fields": (
                    "timestamp",
                    "method",
                    "path",
                    "status_code",
                    "user_agent_details",
                    "querystring",
                    "request_body",
                ),
            },
        ),
        (
            _("Exception Details"),
            {
                "fields": (
                    "exception_type",
                    "exception_message",
                    "traceback",
                ),
            },
        ),
        (
            _("IP Addresses"),
            {
                "fields": (
                    "server_ip",
                    "client_ip",
                ),
            },
        ),
    )
    list_display_links = (
        "method",
        "path",
    )
    search_fields = (
        "status_code",
        "exception_message",
        "client_ip",
        "server_ip",
    )
    list_filter = (
        "method",
        "status_code",
        "path",
        "timestamp",
    )

    @admin.display(description=_("User-Agent Summary"))
    def user_agent_summary(self, obj):
        if not obj.user_agent:
            return None
        return self._get_useragent_summary(parse(obj.user_agent))

    @admin.display(description=_("User-Agent Details"))
    def user_agent_details(self, obj):
        if not obj.user_agent:
            return None
        _parsed_useragent_result = parse(obj.user_agent)

        _details = self._get_useragent_details(_parsed_useragent_result)

        return format_html(
            f"<p>{obj.user_agent}</p>"
            f"<li>{_details["device"]}</li>"
            f"<li>{_details["os"]}</li>"
            f"<li>{_details["user_agent"]}</li>"
        )

    @staticmethod
    def _get_useragent_summary(parsed_useragent_result: Result):
        device_summary = (
            parsed_useragent_result.device.family
            if parsed_useragent_result.device
            else "X"
        )
        os_summary = (
            parsed_useragent_result.os.family if parsed_useragent_result.os else "X"
        )
        user_agent_summary = (
            parsed_useragent_result.user_agent.family
            if parsed_useragent_result.user_agent
            else "X"
        )
        return f"{device_summary}/{os_summary}/{user_agent_summary}"

    def _get_useragent_details(self, parsed_useragent_result: Result):
        device_details = (
            f"Device: {parsed_useragent_result.device.family}"
            f"("
            f"{parsed_useragent_result.device.brand}, {parsed_useragent_result.device.model}"
            f")"
            if parsed_useragent_result.device
            else _("No device data found.")
        )
        os_details = (
            f"OS: {parsed_useragent_result.os.family}"
            f"("
            f"{self._empty_str_if_none(parsed_useragent_result.os.major)}."
            f"{self._empty_str_if_none(parsed_useragent_result.os.minor)}."
            f"{self._empty_str_if_none(parsed_useragent_result.os.patch)}"
            f")"
            if parsed_useragent_result.os
            else _("No OS data found.")
        )
        user_agent_details = (
            f"User-Agent: {parsed_useragent_result.user_agent.family}"
            f"("
            f"{self._empty_str_if_none(parsed_useragent_result.os.major)}."
            f"{self._empty_str_if_none(parsed_useragent_result.os.minor)}."
            f"{self._empty_str_if_none(parsed_useragent_result.os.patch)}"
            f")"
            if parsed_useragent_result.user_agent
            else _("No User-Agent data found.")
        )

        return {
            "device": device_details,
            "os": os_details,
            "user_agent": user_agent_details,
        }

    @staticmethod
    def _empty_str_if_none(value):
        if value is None:
            return ""
        return value

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False



import logging
import threading
import time
from queue import Queue

logger = logging.getLogger("logbox")


class ServerLogInsertThread(threading.Thread):
    _instance = None
    _lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super(ServerLogInsertThread, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        logging_daemon_interval: int,
        logging_daemon_queue_size: int,
        name: str = "logbox_logger_thread",
    ):
        super().__init__(name=name, daemon=True)
        from django_logbox.models import ServerLog

        self._serverlog_model = ServerLog
        self._logging_daemon_interval = logging_daemon_interval
        self._logging_daemon_queue_size = logging_daemon_queue_size
        self._queue = Queue(maxsize=self._logging_daemon_queue_size)

    @staticmethod
    def get_instance():
        return ServerLogInsertThread._instance

    def run(self) -> None:
        while True:
            try:
                time.sleep(self._logging_daemon_interval)
                self._start_bulk_insertion()
            except Exception as e:
                logger.error(f"Error occurred while inserting logs: {e}")

    def put_serverlog(self, data) -> None:
        self._queue.put(self._serverlog_model(**data))
        if self._queue.qsize() >= self._logging_daemon_queue_size:
            logger.debug(
                f"Queue is full({self._queue.qsize()}), starting bulk insertion"
            )
            self._start_bulk_insertion()

    def start(self):
        for t in threading.enumerate():
            if t.name == self.name:
                return
        super().start()
        logger.info(f"Logbox logger thread started: {self.name}")

    def _start_bulk_insertion(self):
        bulk_item = []
        while not self._queue.empty():
            bulk_item.append(self._queue.get())
        if bulk_item:
            self._serverlog_model.objects.bulk_create(bulk_item)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class DjangoLogboxConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_logbox"
    verbose_name = _("Logbox")

    def ready(self):
        from django_logbox.app_settings import settings
        from django_logbox.threading import ServerLogInsertThread

        logbox_thread = ServerLogInsertThread(
            logging_daemon_interval=settings.LOGBOX_SETTINGS["LOGGING_DAEMON_INTERVAL"],
            logging_daemon_queue_size=settings.LOGBOX_SETTINGS[
                "LOGGING_DAEMON_QUEUE_SIZE"
            ],
        )
        logbox_thread.start()
